@@ -12,17 +12,21 @@
 #	See the License for the specific language governing permissions and
 #	limitations under the License.
 
-import itertools
 import sys
 import os
 import json
 import base64
 import logging
 
-from . import utils
+from pouchdb import utils
 
 logger = logging.getLogger(__name__)
 
+#A lot of 'pragma: no cover/branch' statements here. Every path is
+#covered when running python setup.py extensive_test, but that
+#script uses multiple different Python versions and Qt bindings.That is
+#impractical for coverage generation.
+
 def importBestSuitedQtBinding():
 	#first try to import a binding imported by another module already
 	bindings = ["PySide", "PyQt5", "PyQt4"]
@@ -47,19 +51,19 @@ def importBinding(binding):
 	}[binding]()
 
 def importPySide():
-	global QtCore, QtGui, QtWebKit, QtWebKitWidgets
-	from PySide import QtCore, QtGui, QtWebKit
+	global QtCore, QtGui, QtNetwork, QtWebKit, QtWebKitWidgets
+	from PySide import QtCore, QtGui, QtNetwork, QtWebKit
 	QtWebKitWidgets = QtWebKit
 	logger.info("Imported PySide")
 
 def importPyQt5(): #pragma: no cover
-	global QtCore, QtGui, QtWebKit, QtWebKitWidgets
-	from PyQt5 import QtCore, QtWebKitWidgets, QtWebKit, QtWidgets as QtGui
+	global QtCore, QtGui, QtWebKit, QtNetwork, QtWebKitWidgets
+	from PyQt5 import QtCore, QtNetwork, QtWebKitWidgets, QtWebKit, QtWidgets as QtGui
 	logger.info("Imported PyQt5")
 
 def importPyQt4(): #pragma: no cover
-	global QtCore, QtGui, QtWebKit, QtWebKitWidgets
-	from PyQt4 import QtCore, QtGui, QtWebKit
+	global QtCore, QtGui, QtWebKit, QtNetwork, QtWebKitWidgets
+	from PyQt4 import QtCore, QtGui, QtNetwork, QtWebKit
 	QtWebKitWidgets = QtWebKit
 	logger.info("Imported PyQt4")
 
@@ -86,16 +90,6 @@ class Page(QtWebKitWidgets.QWebPage):
 		self._messages = []
 		return r
 
-class AttrAccessDict(dict):
-	def __getattr__(self, key):
-		try:
-			return super(AttrAccessDict, self).__getattr__(key)
-		except AttributeError:
-			try:
-				return self[key]
-			except KeyError, e:
-				raise AttributeError(e)
-
 class Gate(QtCore.QObject):
 	try:
 		Signal = QtCore.Signal
@@ -109,6 +103,17 @@ class JSError(Exception):
 
 		super(JSError, self).__init__(message, *args, **kwargs)
 
+class JSON(object):
+	def __init__(self, data):
+		self.data = data
+
+class Blob(object):
+	def __init__(self, data):
+		self.data = data
+
+	def toJS(self):
+		return "b64toBlob('%s')" % unicode(base64.b64encode(self.data), encoding="ascii")
+
 class JSContext(object):
 	JSError = JSError
 
@@ -126,7 +131,7 @@ class JSContext(object):
 		#statement callback being called first after a Python command.
 		#Otherwise you risk having a promise that has already fired
 		#before said promise is accessable to a Python user.
-		self._evalJsWaiting = True
+		self._evalJsWaiting = False
 
 		#this is used to track functions that are being waited on via
 		#waitUntilCalled.
@@ -162,7 +167,10 @@ class JSContext(object):
 				logger.debug("offlineStoragePath - old: %s, new: %s", currentPath, path)
 				raise ValueError("QWebSettings' offlineStoragePath() property is already set to another value. Impossible to change it since it's a static property.")
 
-		self._reload()
+		#Qt guarantees the html is loaded immeadiately, and since there
+		#aren't any resources, we don't need to block until loaded.
+		self._page.mainFrame().setHtml(self._html, QtCore.QUrl(self._baseUrl))
+		logger.debug("(re)loaded page")
 
 	def getDbSize(self, name):
 		for db in self._page.mainFrame().securityOrigin().databases():
@@ -178,30 +186,20 @@ class JSContext(object):
 
 		"""
 		QtWebKit.QWebSettings.globalSettings().setAttribute(QtWebKit.QWebSettings.DeveloperExtrasEnabled, True)
-		self._i = QtWebKitWidgets.QWebInspector()
-		self._i.setPage(self._page)
+		if not hasattr(self, "_i"): #pragma: no branch
+			self._i = QtWebKitWidgets.QWebInspector()
+			self._i.setPage(self._page)
 		self._i.show()
 		while block and self._i.isVisible():#pragma: no cover
 			self._app.processEvents()
 		self._i.close()
 		QtWebKit.QWebSettings.globalSettings().setAttribute(QtWebKit.QWebSettings.DeveloperExtrasEnabled, False)
 
-	def reset(self):
-		"""Removes all databases and reloads the page."""
-
-		for db in self._page.mainFrame().securityOrigin().databases():
-			QtWebKit.QWebDatabase.removeDatabase(db)
-		logger.debug("Removed all databases")
-		self._reload()
-
-	def _reload(self):
-		#Qt guarantees the html is loaded immeadiately, and since there
-		#aren't any resources, we don't need to block until loaded.
-		self._page.mainFrame().setHtml(self._html, QtCore.QUrl(self._baseUrl))
-		logger.debug("(re)loaded page")
+	def clearCookies(self):
+		self._page.networkAccessManager().setCookieJar(QtNetwork.QNetworkCookieJar())
 
 	def _callbackCalled(self, id, args):
-		if self._evalJsWaiting and id != "_evalJs":
+		if self._evalJsWaiting and id != "evalJs":
 			#schedule to run this function later again - when this
 			#condition is hopefully solved.
 			QtCore.QTimer.singleShot(0, lambda: self._callbackCalled(id, args))
@@ -226,7 +224,7 @@ class JSContext(object):
 		#function conversion
 		with utils.suppress(TypeError, KeyError):
 			if arg["type"] == "_js_returned_function":
-				return lambda *args: self._evalJs("return functions[%s](%s);" % (
+				return lambda *args: self.evalJs("return functions[%s](%s);" % (
 					arg["functionId"],
 					self._toJSArgs(args),
 				))
@@ -239,7 +237,7 @@ class JSContext(object):
 				}
 		#object: recursion
 		with utils.suppress(AttributeError):
-			newArg = AttrAccessDict()
+			newArg = utils.AttrAccessDict()
 			for key, value in arg.iteritems():
 				newArg[key] = self._toPyArg(value)
 			return newArg
@@ -252,7 +250,7 @@ class JSContext(object):
 		return arg
 
 	def newObject(self, id, *args):
-		self._evalJs("""
+		self.evalJs("""
 			objects[{id}] = new {className}({args});
 		""".format(
 			className=self._className,
@@ -268,10 +266,14 @@ class JSContext(object):
 		if callable(arg):
 			self._functions[str(id(arg))] = arg
 			return "createCallback('%s')" % id(arg)
+
 		#'blobs'
-		with utils.suppress(TypeError, KeyError):
-			if arg["type"] == "_py_blob":
-				return "b64toBlob('%s')" % unicode(base64.b64encode(arg["data"]), encoding="ascii")
+		if isinstance(arg, Blob):
+			return arg.toJS()
+
+		#json values can pass through - they are already valid JS.
+		if isinstance(arg, JSON):
+			return arg.data
 
 		#mappings - handle recursively
 		with utils.suppress(AttributeError):
@@ -279,15 +281,13 @@ class JSContext(object):
 			for k, v in arg.iteritems():
 				obj += '"' + k +'":' + self._toJSArg(v) + ","
 			return obj.strip(",") + "}"
-		#TODO: if context.py would ever be used for another project,
-		#adding support for sequences here would be nice.
 
 		#the default
 		return json.dumps(arg)
 
 	def staticProperty(self, propertyName):
 		js = "return %s['%s']" % (self._className, propertyName)
-		return self._evalJs(js)
+		return self.evalJs(js)
 
 	def callStatic(self, funcName, *args):
 		js = "return {className}.{funcName}({args});".format(
@@ -296,7 +296,7 @@ class JSContext(object):
 			funcName=funcName,
 			id=id,
 		)
-		return self._evalJs(js)
+		return self.evalJs(js)
 
 	def call(self, id, funcName, *args):
 		js = "return objects[{id}].{funcName}({args});".format(
@@ -304,7 +304,7 @@ class JSContext(object):
 			funcName=funcName,
 			id=id,
 		)
-		return self._evalJs(js)
+		return self.evalJs(js)
 
 	def waitUntilCalled(self, callbacks):
 		"""Blocks until one of the `callbacks` is called. When passing
@@ -334,19 +334,19 @@ class JSContext(object):
 			#to clean the set if called with multiple callbacks
 			self._waitedOn.discard(callback)
 
-	def _evalJs(self, userJs):
+	def evalJs(self, userJs):
 		logger.debug("evalling JS: %s", userJs.strip())
 		returned = {}
 		def callback(result=None):
 			self._evalJsWaiting = False
 			returned["result"] = result
-		self._functions["_evalJs"] = callback
+		self._functions["evalJs"] = callback
 		js = """
 			var result = (function () {
 				%s
 			}());
-			createCallback('_evalJs')(result);
-		""" % userJs
+			createCallback('evalJs')(result);
+		""".replace("\n", "") % userJs
 		self._evalJsWaiting = True
 		self._page.mainFrame().evaluateJavaScript(js)
 
